import sys
import os
import numpy as np
import tempfile
import threading
import shutil

sys.path.append(os.path.join(os.getenv('JCMROOT'), 'ThirdPartySupport', 'Python'))
import jcmwave

# Definition of the search domain
domain = [
    {'name': 'height', 'type': 'continuous', 'domain': (50,200)}, 
    {'name': 'top_width', 'type': 'continuous', 'domain': (60,200)}, 
    {'name': 'bottom_width', 'type': 'continuous', 'domain': (60,200)}, 
    {'name': 'periodicity', 'type': 'continuous', 'domain': (150,400)}
]

# Definition of constraints
constraints = [
    {'name':'max_bottom_width', 'constraint': 'bottom_width - periodicity + 10'},
    {'name':'max_top_width', 'constraint': 'top_width - periodicity + 10'}
]

# Creation of the study object
study = jcmwave.optimizer.create_study(domain=domain, constraints=constraints,
            name='Optimization of antireflection coating')
study.set_parameters(max_iter=20, num_parallel=2)

# Startup deamon
jcmwave.daemon.startup()
jcmwave.daemon.add_workstation(
    Hostname='localhost',
    Multiplicity=2, NThreads=1)

# Definition of the objective function that is to be minimized,
# i.e. the averaged reflectivity of the meta surface.
def objective(**kwargs):
    keys = kwargs
    keys['slc'] = 150
    
    # The reflectivity is averaged over the wavelengths 600nm 700nm 800nm
    lambdas = [600, 700, 800]
    permittivities = [15.590 + 0.21635j, 14.317 + 0.092101j, 13.646 + 0.048345j]
    
    # Do a parallel lambda-scan
    job_ids = []
    for idx, lambda0 in enumerate(lambdas):
        keys['lambda0'] = lambda0*1e-9
        keys['permittivity'] = permittivities[idx]       
        job_id = jcmwave.solve('project.jcmp', keys,
                               temporary=True)
        job_ids.append(job_id)

    results, logs = jcmwave.daemon.wait(job_ids)
        
    # Sum up reflectivities and their parameter derivatives for each wavelength
    reflectivity = 0
    derivatives = np.zeros(len(domain))
    for idx, lambda0 in enumerate(lambdas):    
        # The Fourier coefficients are stored in the results of the first post process
        FC = results[idx][1]['ElectricFieldStrength'][0]
        # Reflectivity for the current wavelength 
        this_reflectivity = np.sum(np.real(FC)**2 + np.imag(FC)**2)
        # Averaged reflectivity 
        reflectivity += this_reflectivity/len(lambdas)
        for idx2,info in enumerate(domain):
            # Get derivatives of Fourier coefficients from the entries 'd_height', 'd_top_width', ... 
            dFC = results[idx][1]['d_'+info['name']]['ElectricFieldStrength'][0]
            # Derivatives of the reflectivity according to the chain rule
            d_this_reflectivity = np.sum(2*(np.real(FC)*np.real(dFC) +
                                            np.imag(FC)*np.imag(dFC)))
            derivatives[idx2] += d_this_reflectivity/len(lambdas)

    # Create a new observation object and add the objective value
    # and its derivatives
    observation = study.new_observation()
    observation.add(reflectivity)
    for idx,info in enumerate(domain):
        observation.add(derivative=info['name'], value=derivatives[idx])

    return observation

# Definition of a function that is called in each parallel thread in order
# to acquire the objective observation and add it to the study.
def acquire(suggestion):
    observation = objective(**suggestion.kwargs)
    study.add_observation(observation, suggestion.id)

# Run the minimization 
while (not study.is_done()):
    suggestion = study.get_suggestion()
    t = threading.Thread(target=acquire, args=(suggestion,))
    t.daemon = True
    t.start()
    
