import jcmwave
import numpy as np
import numpy.matlib as matlib
from copy import copy
import os

jcmwave.daemon.shutdown()
jcmwave.daemon.add_workstation(Hostname ='localhost', 
                               Multiplicity = 2, 
                               NThreads = 1)

#define default key values
default_keys = {
    'finite_element_degree': 3,
    'maximum_sidelength': 0.1,
    'radius': 0.3
    }

#create resultbag. Results will be stored in file ./resultbag.db
resultbag = jcmwave.Resultbag('resultbag.db')

#define set of keys for a multi-dimensional parameter scan
sidelengths = np.linspace(0.5, 0.1, 4)
radii = np.linspace(0.3, 0.5, 8)
scan1,scan2 = np.meshgrid(sidelengths, radii)
keyset = matlib.tile({},scan1.shape) 
for ii in np.ndindex(*scan1.shape):
    keyset[ii] = copy(default_keys)
    keyset[ii]['maximum_sidelength'] = scan1[ii]
    keyset[ii]['radius'] = scan2[ii]

#run the parameter scan
job_ids = []

# loop over radius values
for ii in np.ndindex(*keyset.shape):
   keys = keyset[ii]
   job_id = jcmwave.solve('mie2D.jcmp', keys = keys, 
                          temporary = True,
                          resultbag = resultbag)
   job_ids.append(job_id)

# jcm_deamon_wait adds all results to the resultbag
jcmwave.daemon.wait(job_ids, resultbag = resultbag)

# now we collect the computed scattering cross section 
# as in the the sequential example 
scattering_cross_section_scan = np.zeros(keyset.shape)
for ii in np.ndindex(*keyset.shape):
   keys = keyset[ii]
   result = resultbag.get_result(keys)
   scs = result[1]['ElectromagneticFieldEnergyFlux'][0][0].real
   scattering_cross_section_scan[ii] = scs
 
# plot scattering cross section against rod radius
import matplotlib.pyplot as plt
handles = []
for data in zip(*scattering_cross_section_scan):
   handle, = plt.plot(radii, data, '-+', linewidth=2, markersize=14)
   handles.append(handle)
plt.legend(handles, ['sidelength %.1f' % s for s in sidelengths])
plt.xlabel('radius [$\mu$ m]', fontsize=14)
plt.ylabel('integral scattering cross section', fontsize=14)
plt.axis('tight')
plt.show()
